/*----- PROTECTED REGION ID(MagnetType.cpp) ENABLED START -----*/
static const char *RcsId = "$Id:  $";
//=============================================================================
//
// file :        MagnetType.cpp
//
// description : C++ source for the MagnetType class and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               MagnetType are implemented in this file.
//
// project :     MagnetType
//
// This file is part of Tango device class.
//
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
//
// $Author:  $
// Copyright (C): 2004,2005,2006,2007,2008,2009,2009,2010,2011,2012,2013,2014,2015
//                European Synchrotron Radiation Facility
//                BP 220, Grenoble 38043
//                France
//
// $Revision:  $
// $Date:  $
//
// $HeadURL:  $
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include <MagnetType.h>
#include <MagnetTypeClass.h>

#include <future>

#include <iomanip>

/*----- PROTECTED REGION END -----*/	//	MagnetType.cpp

/**
 *  MagnetType class description:
 *    This class is for the ESRF EBS storage ring simulator.
 *    It is the interface for a whole set of magnets from the same type
 *    (all quadrupoles, all sextupoles,...). This class is supposed to
 *    be used by control loop software
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name             |  Method name
//================================================================
//  State                    |  dev_state
//  Status                   |  Inherited (no method)
//  Reset                    |  reset
//  ResetResonanceStrengths  |  reset_resonance_strengths
//  SetpointCheck            |  setpoint_check
//================================================================

//================================================================
//  Attributes managed are:
//================================================================
//  DesignStrengthFile            |  Tango::DevString	Scalar
//  CorrectionStrengthsSteps      |  Tango::DevLong	Scalar
//  Strengths                     |  Tango::DevDouble	Spectrum  ( max = 610)
//  StrengthsMSMM                 |  Tango::DevDouble	Spectrum  ( max = 4)
//  StrengthsReadPoint            |  Tango::DevDouble	Spectrum  ( max = 610)
//  TotalCorrectionStrengths      |  Tango::DevDouble	Spectrum  ( max = 610)
//  TotalCorrectionStrengthsMSMM  |  Tango::DevDouble	Spectrum  ( max = 4)
//  CorrectionStrengths           |  Tango::DevDouble	Spectrum  ( max = 610)
//  CorrectionStrengthsMSMM       |  Tango::DevDouble	Spectrum  ( max = 4)
//  ResonanceStrengths            |  Tango::DevDouble	Spectrum  ( max = 610)
//  ResonanceStrengthsMSMM        |  Tango::DevDouble	Spectrum  ( max = 4)
//  MagnetNames                   |  Tango::DevString	Spectrum  ( max = 610)
//  MagnetStates                  |  Tango::DevState	Spectrum  ( max = 620)
//  RequireCycling                |  Tango::DevBoolean	Spectrum  ( max = 620)
//  FrozenMagnets                 |  Tango::DevString	Spectrum  ( max = 620)
//  IsFrozen                      |  Tango::DevBoolean	Spectrum  ( max = 620)
//================================================================

namespace MagnetType_ns
{
/*----- PROTECTED REGION ID(MagnetType::namespace_starting) ENABLED START -----*/

//	static initializations

/*----- PROTECTED REGION END -----*/	//	MagnetType::namespace_starting

//--------------------------------------------------------
/**
 *	Method     : MagnetType::MagnetType()
 *	Description: Constructors for a Tango device
 *                implementing the classMagnetType
 */
//--------------------------------------------------------
MagnetType::MagnetType(Tango::DeviceClass *cl, std::string &s)
 : TANGO_BASE_CLASS(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(MagnetType::constructor_1) ENABLED START -----*/
	init_device();

	/*----- PROTECTED REGION END -----*/	//	MagnetType::constructor_1
}
//--------------------------------------------------------
MagnetType::MagnetType(Tango::DeviceClass *cl, const char *s)
 : TANGO_BASE_CLASS(cl, s)
{
	/*----- PROTECTED REGION ID(MagnetType::constructor_2) ENABLED START -----*/
	init_device();

	/*----- PROTECTED REGION END -----*/	//	MagnetType::constructor_2
}
//--------------------------------------------------------
MagnetType::MagnetType(Tango::DeviceClass *cl, const char *s, const char *d)
 : TANGO_BASE_CLASS(cl, s, d)
{
	/*----- PROTECTED REGION ID(MagnetType::constructor_3) ENABLED START -----*/
	init_device();

	/*----- PROTECTED REGION END -----*/	//	MagnetType::constructor_3
}
//--------------------------------------------------------
MagnetType::~MagnetType()
{
	delete_device();
}

//--------------------------------------------------------
/**
 *	Method     : MagnetType::delete_device()
 *	Description: will be called at device destruction or at init command
 */
//--------------------------------------------------------
void MagnetType::delete_device()
{
	DEBUG_STREAM << "MagnetType::delete_device() " << device_name << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::delete_device) ENABLED START -----*/

    delete grp;

    delete [] attr_CorrectionStrengths_write;
    delete [] attr_ResonanceStrengths_write;
    delete [] attr_Strengths_write;
    delete [] attr_TotalCorrectionStrengths_write;
    delete [] attr_IsFrozen_write;

  if(attr_MagnetNames_read!=nullptr) {
    for (int ind = 0; ind < nb_magnet; ind++)
      CORBA::string_free(attr_MagnetNames_read[ind]);
    delete[] attr_MagnetNames_read;
  }

	/*----- PROTECTED REGION END -----*/	//	MagnetType::delete_device
	delete[] attr_DesignStrengthFile_read;
	delete[] attr_CorrectionStrengthsSteps_read;
	delete[] attr_Strengths_read;
	delete[] attr_StrengthsReadPoint_read;
	delete[] attr_TotalCorrectionStrengths_read;
	delete[] attr_CorrectionStrengths_read;
	delete[] attr_ResonanceStrengths_read;
	delete[] attr_MagnetStates_read;
	delete[] attr_RequireCycling_read;
	delete[] attr_IsFrozen_read;
}

//--------------------------------------------------------
/**
 *	Method     : MagnetType::init_device()
 *	Description: will be called at device initialization.
 */
//--------------------------------------------------------
void MagnetType::init_device()
{
	DEBUG_STREAM << "MagnetType::init_device() create device " << device_name << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::init_device_before) ENABLED START -----*/

    grp = nullptr;
    attr_CorrectionStrengths_write = new Tango::DevDouble[MAX_MAGNET_NB];
    attr_ResonanceStrengths_write = new Tango::DevDouble[MAX_MAGNET_NB];
    attr_Strengths_write = new Tango::DevDouble[MAX_MAGNET_NB];
    attr_TotalCorrectionStrengths_write = new Tango::DevDouble[MAX_MAGNET_NB];
    attr_IsFrozen_write = new Tango::DevBoolean[MAX_MAGNET_NB];
    attr_MagnetNames_read = nullptr;

    corr_msmm.assign(4,0.0);
    reso_msmm.assign(4,0.0);
    str_msmm.assign(4,0.0);
    total_msmm.assign(4,0.0);

    /*----- PROTECTED REGION END -----*/	//	MagnetType::init_device_before


	//	Get the device properties from database
	get_device_property();

	attr_DesignStrengthFile_read = new Tango::DevString[1];
	attr_CorrectionStrengthsSteps_read = new Tango::DevLong[1];
	attr_Strengths_read = new Tango::DevDouble[610];
	attr_StrengthsReadPoint_read = new Tango::DevDouble[610];
	attr_TotalCorrectionStrengths_read = new Tango::DevDouble[610];
	attr_CorrectionStrengths_read = new Tango::DevDouble[610];
	attr_ResonanceStrengths_read = new Tango::DevDouble[610];
	attr_MagnetStates_read = new Tango::DevState[620];
	attr_RequireCycling_read = new Tango::DevBoolean[620];
	attr_IsFrozen_read = new Tango::DevBoolean[620];
	//	No longer if mandatory property not set.
	if (mandatoryNotDefined)
		return;

	/*----- PROTECTED REGION ID(MagnetType::init_device) ENABLED START -----*/

  *attr_CorrectionStrengthsSteps_read = 1;
  Tango::WAttribute &att_step = get_device_attr()->get_w_attr_by_name("CorrectionStrengthsSteps");
  att_step.set_write_value(1);

  // Check  the magnet type

  transform(magnetType.begin(),magnetType.end(),magnetType.begin(),::tolower);
  if (magnetType == "quadrupole")
  {
    mag_type = MagnetType::MagType::QUAD;
    nb_magnet = QUAD_NB + DQ_NB;
  }
  else if (magnetType == "sextupole")
  {
    mag_type = MagnetType::MagType::SEXTU;
    nb_magnet = SEXTU_NB;

  }
  else if (magnetType == "octupole")
  {
    mag_type = MagnetType::MagType::OCTU;
    nb_magnet = OCTU_NB;
  }

  if( magnetNames.size() != nb_magnet ) {
    Tango::Except::throw_exception(
            "ErrorInit",
            "Wrong MagnetNames property definition (" + std::to_string(nb_magnet) + " items expected",
            "MagnetType::init_device"
    );
  }

  // Create a group with all devices for the magnet type.
  Tango::GroupReply::enable_exception(false);
  grp = new Tango::Group("Magnets");
  grp->add(magnetNames);
  grp->set_timeout_millis(4000);

  // Check import
  vector<string> device_list = grp->get_device_list();
  for (auto &elem : device_list)
    transform(elem.begin(), elem.end(), elem.begin(), ::tolower);

  for (auto &elem : device_list)
    DEBUG_STREAM << "device in group = " << elem << endl;

  if( device_list.size() != nb_magnet ) {
    Tango::Except::throw_exception(
            "ErrorInit",
            "Wrong MagnetNames property definition (one or more item is invalid)",
            "MagnetType::init_device"
    );
  }

  // Init the magnet names attribute

  attr_MagnetNames_read = new Tango::DevString [nb_magnet];
  for(size_t i=0;i<magnetNames.size();i++) {
    attr_MagnetNames_read[i] = Tango::string_dup(magnetNames[i]);
    attr_MagnetStates_read[i] = Tango::UNKNOWN;
  }

  // Set attributes unit: quad -> m-1, sextu -> m-2, octu -> m-3

  set_att_unit("CorrectionStrengths");
  set_att_unit("CorrectionStrengthsMSMM");

  set_att_unit("ResonanceStrengths");
  set_att_unit("ResonanceStrengthsMSMM");

  set_att_unit("Strengths");
  set_att_unit("StrengthsMSMM");

  set_att_unit("TotalCorrectionStrengths");
  set_att_unit("TotalCorrectionStrengthsMSMM");

  // Some init

  design_file_name = "?";
  *attr_DesignStrengthFile_read = const_cast<char *>(design_file_name.c_str());

  set_state(Tango::ON);
  set_status(Tango::StatusNotSet);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::init_device
}

//--------------------------------------------------------
/**
 *	Method     : MagnetType::get_device_property()
 *	Description: Read database to initialize property data members.
 */
//--------------------------------------------------------
void MagnetType::get_device_property()
{
	/*----- PROTECTED REGION ID(MagnetType::get_device_property_before) ENABLED START -----*/

	//	Initialize property data members

	/*----- PROTECTED REGION END -----*/	//	MagnetType::get_device_property_before

	mandatoryNotDefined = false;

	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("MagnetType"));
	dev_prop.push_back(Tango::DbDatum("MagnetNames"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);

		//	get instance on MagnetTypeClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		MagnetTypeClass	*ds_class =
			(static_cast<MagnetTypeClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize MagnetType from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  magnetType;
		else {
			//	Try to initialize MagnetType from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  magnetType;
		}
		//	And try to extract MagnetType value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  magnetType;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize MagnetNames from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  magnetNames;
		else {
			//	Try to initialize MagnetNames from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  magnetNames;
		}
		//	And try to extract MagnetNames value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  magnetNames;

	}

	/*----- PROTECTED REGION ID(MagnetType::get_device_property_after) ENABLED START -----*/

	//	Check device property data members init

	/*----- PROTECTED REGION END -----*/	//	MagnetType::get_device_property_after
}
//--------------------------------------------------------
/**
 *	Method     : MagnetType::check_mandatory_property()
 *	Description: For mandatory properties check if defined in database.
 */
//--------------------------------------------------------
void MagnetType::check_mandatory_property(Tango::DbDatum &class_prop, Tango::DbDatum &dev_prop)
{
	//	Check if all properties are empty
	if (class_prop.is_empty() && dev_prop.is_empty())
	{
		TangoSys_OMemStream	tms;
		tms << std::endl <<"Property \'" << dev_prop.name;
		if (Tango::Util::instance()->_UseDb==true)
			tms << "\' is mandatory but not defined in database";
		else
			tms << "\' is mandatory but cannot be defined without database";
		append_status(tms.str());
		mandatoryNotDefined = true;
		/*----- PROTECTED REGION ID(MagnetType::check_mandatory_property) ENABLED START -----*/
		cerr << tms.str() << " for " << device_name << endl;

		/*----- PROTECTED REGION END -----*/	//	MagnetType::check_mandatory_property
	}
}


//--------------------------------------------------------
/**
 *	Method     : MagnetType::always_executed_hook()
 *	Description: method always executed before any command is executed
 */
//--------------------------------------------------------
void MagnetType::always_executed_hook()
{
	DEBUG_STREAM << "MagnetType::always_executed_hook()  " << device_name << std::endl;
	if (mandatoryNotDefined)
	{
		Tango::Except::throw_exception(
					(const char *)"PROPERTY_NOT_SET",
					get_status().c_str(),
					(const char *)"MagnetType::always_executed_hook()");
	}
	/*----- PROTECTED REGION ID(MagnetType::always_executed_hook) ENABLED START -----*/

	//	code always executed before all requests

	/*----- PROTECTED REGION END -----*/	//	MagnetType::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method     : MagnetType::read_attr_hardware()
 *	Description: Hardware acquisition for attributes
 */
//--------------------------------------------------------
void MagnetType::read_attr_hardware(TANGO_UNUSED(std::vector<long> &attr_list))
{
	DEBUG_STREAM << "MagnetType::read_attr_hardware(std::vector<long> &attr_list) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_attr_hardware) ENABLED START -----*/

    if (design_file_name[0] == '?')
    {
        Tango::WAttribute &watt = get_device_attr()->get_w_attr_by_name("DesignStrengthFile");
        char *dsf_ptr;
        watt.get_write_value(dsf_ptr);
        cout << "ptr = " << dsf_ptr << endl;
        design_file_name = dsf_ptr;
        *attr_DesignStrengthFile_read = const_cast<char *>(design_file_name.c_str());
    }

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_attr_hardware
}
//--------------------------------------------------------
/**
 *	Method     : MagnetType::write_attr_hardware()
 *	Description: Hardware writing for attributes
 */
//--------------------------------------------------------
void MagnetType::write_attr_hardware(TANGO_UNUSED(std::vector<long> &attr_list))
{
	DEBUG_STREAM << "MagnetType::write_attr_hardware(std::vector<long> &attr_list) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::write_attr_hardware) ENABLED START -----*/

	//	Add your own code

	/*----- PROTECTED REGION END -----*/	//	MagnetType::write_attr_hardware
}

//--------------------------------------------------------
/**
 *	Read attribute DesignStrengthFile related method
 *	Description: Design strength file name
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void MagnetType::read_DesignStrengthFile(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_DesignStrengthFile(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_DesignStrengthFile) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_DesignStrengthFile_read);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_DesignStrengthFile
}
//--------------------------------------------------------
/**
 *	Write attribute DesignStrengthFile related method
 *	Description: Design strength file name
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void MagnetType::write_DesignStrengthFile(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MagnetType::write_DesignStrengthFile(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevString	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(MagnetType::write_DesignStrengthFile) ENABLED START -----*/

  // Try to open and read the design strength file (CSV format)
  vector<double> magnet_vals;

  string f_name(w_val);

  ifstream infile(f_name);
  if (!infile.is_open()) {
    stringstream ss;
    ss << "Not able to open design strength file " << f_name;
    Tango::Except::throw_exception("MagnetType_CantOpenFile",ss.str(),"MagnetType::write_DesignStrengthFile()");
  } else {

    string line;
    size_t li = 1;

    try
    {
      while (getline(infile, line))
      {
        stringstream strstr(line);
        string word;
        string name;
        double val = 0.0;
        int ctr = 0;

        while (getline(strstr,word, ','))
        {
          if (ctr == 0)
          {
            name = word;
            transform(name.begin(),name.end(),name.begin(),::tolower);
          }
          else
            val = stod(word);
          ctr++;
        }

        switch(mag_type)
        {
          case MagType::QUAD:
            if (name.find("/m-q") != string::npos or name.find("m-dq") != string::npos)
              magnet_vals.push_back(val);
            break;

          case MagType::SEXTU:
            if (name.find("/m-s") != string::npos)
              magnet_vals.push_back(val);
            break;

          case MagType::OCTU:
            if (name.find("/m-o") != string::npos)
              magnet_vals.push_back(val);
            break;
        }
        li++;
      }
    }
    catch (...)
    {
      stringstream ss;
      ss << "Error when reading file " << f_name << " at line " << li;
      Tango::Except::throw_exception("MagnetType_WrongFile",ss.str(),"MagnetType::write_DesignStrengthFile()");
    }
  }

  // Check number of read values
  if (magnet_vals.size() != nb_magnet)
    Tango::Except::throw_exception(
            "MagnetType_WrongFile",
            "Wrong number of elt in file (" + std::to_string(nb_magnet) + " expected, but got " + std::to_string(magnet_vals.size()) + ")",
            "MagnetType::write_DesignStrengthFile()"
    );

  design_strength = magnet_vals;

  // Send data
  Tango::GroupCmdReplyList grl = grp->command_inout("SetDesignStrength",magnet_vals);

  string err_mess;
  for (size_t loop = 0;loop < grl.size();loop++)
  {
    if (grl[loop].has_failed())
    {
      const Tango::DevErrorList &del = grl[loop].get_err_stack();
      string dev_name = grp->get_device(loop + 1)->name();
      string tmp;
      tmp = "Error for device " + dev_name + ": " + del[0].desc.in();
      if (err_mess.empty() == false)
        err_mess.append("\n");
      err_mess.append(tmp);
    }
  }

  if (!err_mess.empty())
    Tango::Except::throw_exception("MagnetType_CantWriteStrength",err_mess,"MagnetType::write_DesignStrengthFile()");

  design_file_name = f_name;
  *attr_DesignStrengthFile_read = const_cast<char *>(design_file_name.c_str());

	/*----- PROTECTED REGION END -----*/	//	MagnetType::write_DesignStrengthFile
}
//--------------------------------------------------------
/**
 *	Read attribute CorrectionStrengthsSteps related method
 *	Description: Number of steps in which a new Correction Strenths must be applied
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void MagnetType::read_CorrectionStrengthsSteps(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_CorrectionStrengthsSteps(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_CorrectionStrengthsSteps) ENABLED START -----*/

    attr.set_value(attr_CorrectionStrengthsSteps_read);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_CorrectionStrengthsSteps
}
//--------------------------------------------------------
/**
 *	Write attribute CorrectionStrengthsSteps related method
 *	Description: Number of steps in which a new Correction Strenths must be applied
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void MagnetType::write_CorrectionStrengthsSteps(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MagnetType::write_CorrectionStrengthsSteps(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(MagnetType::write_CorrectionStrengthsSteps) ENABLED START -----*/

    *attr_CorrectionStrengthsSteps_read = w_val;

	/*----- PROTECTED REGION END -----*/	//	MagnetType::write_CorrectionStrengthsSteps
}
//--------------------------------------------------------
/**
 *	Read attribute Strengths related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 610
 */
//--------------------------------------------------------
void MagnetType::read_Strengths(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_Strengths(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_Strengths) ENABLED START -----*/

  Tango::GroupAttrReplyList arl = grp->read_attribute("Strength");

  for (size_t i = 0;i < arl.size();i++) {
    if (arl[i].has_failed()) {
      attr_Strengths_read[i] = NAN;
      attr_Strengths_write[i] = NAN;
    } else {
      if (arl[i].get_data().get_quality() == Tango::ATTR_INVALID) {
        attr_Strengths_read[i] = NAN;
        attr_Strengths_write[i] = NAN;
      } else {
        vector<double> v;
        arl[i] >> v;
        attr_Strengths_read[i] = v[0];
        attr_Strengths_write[i] = v[1];
      }
    }
  }

  compute_MSMM(attr_Strengths_write,nb_magnet,str_msmm);
  attr.set_value(attr_Strengths_write,nb_magnet);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_Strengths
}
//--------------------------------------------------------
/**
 *	Read attribute StrengthsMSMM related method
 *	Description: Strengths spectrum attribute Mean, Std, Min and Max
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4
 */
//--------------------------------------------------------
void MagnetType::read_StrengthsMSMM(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_StrengthsMSMM(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_StrengthsMSMM) ENABLED START -----*/

    attr.set_value(str_msmm.data(), 4);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_StrengthsMSMM
}
//--------------------------------------------------------
/**
 *	Read attribute StrengthsReadPoint related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 610
 */
//--------------------------------------------------------
void MagnetType::read_StrengthsReadPoint(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_StrengthsReadPoint(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_StrengthsReadPoint) ENABLED START -----*/

  Tango::GroupAttrReplyList arl = grp->read_attribute("Strength");

  for (size_t i = 0;i < arl.size();i++) {
    if (arl[i].has_failed()) {
      attr_StrengthsReadPoint_read[i] = NAN;
    } else {
      if (arl[i].get_data().get_quality() == Tango::ATTR_INVALID) {
        attr_StrengthsReadPoint_read[i] = NAN;
      } else {
        double v;
        arl[i] >> v;
        attr_StrengthsReadPoint_read[i] = v;
      }
    }
  }
  attr.set_value(attr_StrengthsReadPoint_read,nb_magnet);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_StrengthsReadPoint
}
//--------------------------------------------------------
/**
 *	Read attribute TotalCorrectionStrengths related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 610
 */
//--------------------------------------------------------
void MagnetType::read_TotalCorrectionStrengths(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_TotalCorrectionStrengths(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_TotalCorrectionStrengths) ENABLED START -----*/

  vector<string> attNames{"ResonanceStrength","CorrectionStrength"};
  Tango::GroupAttrReplyList arl = grp->read_attributes(attNames);

  for (size_t i = 0;i < arl.size();i+=2) {
    if (arl[i].has_failed() || arl[i + 1].has_failed()) {
      attr_TotalCorrectionStrengths_read[i/2] = NAN;
      attr_TotalCorrectionStrengths_write[i/2] = NAN;
    } else {
      if (arl[i].get_data().get_quality() == Tango::ATTR_INVALID ||
          arl[i + 1].get_data().get_quality() == Tango::ATTR_INVALID) {
        attr_TotalCorrectionStrengths_read[i/2] = NAN;
        attr_TotalCorrectionStrengths_write[i/2] = NAN;
      } else {
        vector<double> v1;
        vector<double> v2;
        arl[i] >> v1;
        arl[i + 1] >> v2;
        attr_TotalCorrectionStrengths_read[i/2] = v1[0] + v2[0];
        attr_TotalCorrectionStrengths_write[i/2] = v1[1] + v2[1];
      }
    }
  }

  compute_MSMM(attr_TotalCorrectionStrengths_write,nb_magnet,total_msmm);
  attr.set_value(attr_TotalCorrectionStrengths_write,nb_magnet);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_TotalCorrectionStrengths
}
//--------------------------------------------------------
/**
 *	Read attribute TotalCorrectionStrengthsMSMM related method
 *	Description: TotalCorrectionStrengths spectrum attribute Mean, Std, Min and Max
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4
 */
//--------------------------------------------------------
void MagnetType::read_TotalCorrectionStrengthsMSMM(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_TotalCorrectionStrengthsMSMM(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_TotalCorrectionStrengthsMSMM) ENABLED START -----*/

    attr.set_value(total_msmm.data(), 4);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_TotalCorrectionStrengthsMSMM
}
//--------------------------------------------------------
/**
 *	Read attribute CorrectionStrengths related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 610
 */
//--------------------------------------------------------
void MagnetType::read_CorrectionStrengths(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_CorrectionStrengths(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_CorrectionStrengths) ENABLED START -----*/

  Tango::GroupAttrReplyList arl = grp->read_attribute("CorrectionStrength");

  for (size_t i = 0;i < arl.size();i++) {
    if (arl[i].has_failed()) {
      attr_CorrectionStrengths_read[i] = NAN;
      attr_CorrectionStrengths_write[i] = NAN;
    } else {
      if (arl[i].get_data().get_quality() == Tango::ATTR_INVALID) {
        attr_CorrectionStrengths_read[i] = NAN;
        attr_CorrectionStrengths_write[i] = NAN;
      } else {
        vector<double> v;
        arl[i] >> v;
        attr_CorrectionStrengths_read[i] = v[0];
        attr_CorrectionStrengths_write[i] = v[1];
      }
    }
  }

  compute_MSMM(attr_CorrectionStrengths_write,nb_magnet,corr_msmm);
  (static_cast<Tango::WAttribute &>(attr)).set_write_value(attr_CorrectionStrengths_write,nb_magnet);
  attr.set_value(attr_CorrectionStrengths_read,nb_magnet);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_CorrectionStrengths
}
//--------------------------------------------------------
/**
 *	Write attribute CorrectionStrengths related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 610
 */
//--------------------------------------------------------
void MagnetType::write_CorrectionStrengths(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MagnetType::write_CorrectionStrengths(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve number of write values
	int	w_length = attr.get_write_value_length();

	//	Retrieve pointer on write values (Do not delete !)
	const Tango::DevDouble	*w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(MagnetType::write_CorrectionStrengths) ENABLED START -----*/

  // Check number of received data
  if ((size_t)w_length != nb_magnet)
    Tango::Except::throw_exception(
            "WrongArg",
            "Wrong number of received data (" + std::to_string(nb_magnet) + " expected but got " + std::to_string(w_length) + ")",
            "MagnetType::write_CorrectionStrengths"
    );

  int nb_steps =  attr_CorrectionStrengthsSteps_read[0];

  if( nb_steps>1 ) {

    double startStrength[nb_magnet];
    double deltaStrength[nb_magnet];

    // Get starting point
    Tango::GroupAttrReplyList arl = grp->read_attribute("CorrectionStrength");

    for (size_t i = 0; i < arl.size(); i++) {
      if (arl[i].has_failed()) {
        startStrength[i] = NAN;
      } else {
        if (arl[i].get_data().get_quality() == Tango::ATTR_INVALID) {
          startStrength[i] = NAN;
        } else {
          vector<double> v;
          arl[i] >> v;
          startStrength[i] = v[1];
        }
      }
    }

    double set[nb_magnet];
    for (size_t i = 0; i < arl.size(); i++) {
      deltaStrength[i] = (w_val[i] - startStrength[i]) / (double) (nb_steps + 1);
      set[i] = w_val[i] + deltaStrength[i];
    }

    for(size_t i = 0; i<nb_steps;i++) {
      for (size_t j = 0; j < nb_magnet; j++)
        set[i] += deltaStrength[i];
      write_attribute<double>("CorrectionStrength",set,nb_magnet);
      sleep(1);
    }

  } else {

    write_attribute<double>("CorrectionStrength",(double *)w_val,w_length);

  }

	/*----- PROTECTED REGION END -----*/	//	MagnetType::write_CorrectionStrengths
}
//--------------------------------------------------------
/**
 *	Read attribute CorrectionStrengthsMSMM related method
 *	Description: CorrectionStrengths spectrum attribute Mean, Std, Min and Max
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4
 */
//--------------------------------------------------------
void MagnetType::read_CorrectionStrengthsMSMM(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_CorrectionStrengthsMSMM(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_CorrectionStrengthsMSMM) ENABLED START -----*/

    attr.set_value(corr_msmm.data(), 4);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_CorrectionStrengthsMSMM
}
//--------------------------------------------------------
/**
 *	Read attribute ResonanceStrengths related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 610
 */
//--------------------------------------------------------
void MagnetType::read_ResonanceStrengths(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_ResonanceStrengths(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_ResonanceStrengths) ENABLED START -----*/

  Tango::GroupAttrReplyList arl = grp->read_attribute("ResonanceStrength");

  for (size_t i = 0;i < arl.size();i++) {
    if (arl[i].has_failed()) {
      attr_ResonanceStrengths_read[i] = NAN;
      attr_ResonanceStrengths_write[i] = NAN;
    } else {
      if (arl[i].get_data().get_quality() == Tango::ATTR_INVALID) {
        attr_ResonanceStrengths_read[i] = NAN;
        attr_ResonanceStrengths_write[i] = NAN;
      } else {
        vector<double> v;
        arl[i] >> v;
        attr_ResonanceStrengths_read[i] = v[0];
        attr_ResonanceStrengths_write[i] = v[1];
      }
    }
  }

  compute_MSMM(attr_ResonanceStrengths_write,nb_magnet,reso_msmm);
  (static_cast<Tango::WAttribute &>(attr)).set_write_value(attr_ResonanceStrengths_write,nb_magnet);
  attr.set_value(attr_ResonanceStrengths_read,nb_magnet);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_ResonanceStrengths
}
//--------------------------------------------------------
/**
 *	Write attribute ResonanceStrengths related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 610
 */
//--------------------------------------------------------
void MagnetType::write_ResonanceStrengths(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MagnetType::write_ResonanceStrengths(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve number of write values
	int	w_length = attr.get_write_value_length();

	//	Retrieve pointer on write values (Do not delete !)
	const Tango::DevDouble	*w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(MagnetType::write_ResonanceStrengths) ENABLED START -----*/

  // Check number of received data
  if ((size_t)w_length != nb_magnet)
    Tango::Except::throw_exception(
            "WrongArg",
            "Wrong number of received data (" + std::to_string(nb_magnet) + " expected but got " + std::to_string(w_length) + ")",
            "MagnetType::write_ResonanceStrengths"
    );

  write_attribute<double>("ResonanceStrength",(double *)w_val,w_length);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::write_ResonanceStrengths
}
//--------------------------------------------------------
/**
 *	Read attribute ResonanceStrengthsMSMM related method
 *	Description: ResonanceStrengths spectrum attribute Mean, Std, Min and Max
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4
 */
//--------------------------------------------------------
void MagnetType::read_ResonanceStrengthsMSMM(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_ResonanceStrengthsMSMM(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_ResonanceStrengthsMSMM) ENABLED START -----*/

    attr.set_value(reso_msmm.data(), 4);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_ResonanceStrengthsMSMM
}
//--------------------------------------------------------
/**
 *	Read attribute MagnetNames related method
 *	Description: Level 1 magnet device names
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Spectrum max = 610
 */
//--------------------------------------------------------
void MagnetType::read_MagnetNames(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_MagnetNames(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_MagnetNames) ENABLED START -----*/

  attr.set_value(attr_MagnetNames_read,nb_magnet);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_MagnetNames
}
//--------------------------------------------------------
/**
 *	Read attribute MagnetStates related method
 *
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Spectrum max = 620
 */
//--------------------------------------------------------
void MagnetType::read_MagnetStates(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_MagnetStates(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_MagnetStates) ENABLED START -----*/

  attr.set_value(attr_MagnetStates_read,nb_magnet);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_MagnetStates
}
//--------------------------------------------------------
/**
 *	Read attribute RequireCycling related method
 *	Description: Boolean set for each magnets which requires cycling (in beam propagation order)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Spectrum max = 620
 */
//--------------------------------------------------------
void MagnetType::read_RequireCycling(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_RequireCycling(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_RequireCycling) ENABLED START -----*/

  Tango::GroupAttrReplyList arl = grp->read_attribute("RequireCycling");
  for (size_t i = 0;i < arl.size();i++) {
    if (arl[i].has_failed()) {
      attr_RequireCycling_read[i] = false;
    } else {
      if (arl[i].get_data().get_quality() == Tango::ATTR_INVALID) {
        attr_RequireCycling_read[i] = false;
      } else {
        arl[i] >> attr_RequireCycling_read[i];
      }
    }
  }
  attr.set_value(attr_RequireCycling_read,nb_magnet);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_RequireCycling
}
//--------------------------------------------------------
/**
 *	Read attribute FrozenMagnets related method
 *	Description: List of disabled magnets
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Spectrum max = 620
 */
//--------------------------------------------------------
void MagnetType::read_FrozenMagnets(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_FrozenMagnets(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_FrozenMagnets) ENABLED START -----*/

  vector<size_t> frozen_inds;

  Tango::GroupAttrReplyList arl = grp->read_attribute("Frozen");
  for (size_t i = 0;i < arl.size();i++) {
    if (arl[i].has_failed()) {
      attr_RequireCycling_read[i] = false;
    } else {
      if (arl[i].get_data().get_quality() == Tango::ATTR_INVALID) {
        attr_RequireCycling_read[i] = false;
      } else {
        Tango::DevBoolean b;
        arl[i] >> b;
        if(b) frozen_inds.emplace_back(i);
      }
    }
  }

  attr_FrozenMagnets_read = new Tango::DevString [frozen_inds.size()];
  for(size_t i=0;i<frozen_inds.size();i++)
    attr_FrozenMagnets_read[i] = Tango::string_dup(attr_MagnetNames_read[i]);
   attr.set_value(attr_FrozenMagnets_read,frozen_inds.size(),0,true);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_FrozenMagnets
}
//--------------------------------------------------------
/**
 *	Read attribute IsFrozen related method
 *	Description: Boolean set for each disabled magnets (in beam propagation order)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Spectrum max = 620
 */
//--------------------------------------------------------
void MagnetType::read_IsFrozen(Tango::Attribute &attr)
{
	DEBUG_STREAM << "MagnetType::read_IsFrozen(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::read_IsFrozen) ENABLED START -----*/

  Tango::GroupAttrReplyList arl = grp->read_attribute("Frozen");

  for (size_t i = 0;i < arl.size();i++) {
    if (arl[i].has_failed()) {
      attr_IsFrozen_read[i] = false;
      attr_IsFrozen_write[i] = false;
    } else {
      if (arl[i].get_data().get_quality() == Tango::ATTR_INVALID) {
        attr_IsFrozen_read[i] = false;
        attr_IsFrozen_write[i] = false;
      } else {
        vector<Tango::DevBoolean> v;
        arl[i] >> v;
        attr_IsFrozen_read[i] = v[0];
        attr_IsFrozen_write[i] = v[1];
      }
    }
  }

  (static_cast<Tango::WAttribute &>(attr)).set_write_value(attr_IsFrozen_write,nb_magnet);
  attr.set_value(attr_IsFrozen_read,nb_magnet);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::read_IsFrozen
}
//--------------------------------------------------------
/**
 *	Write attribute IsFrozen related method
 *	Description: Boolean set for each disabled magnets (in beam propagation order)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Spectrum max = 620
 */
//--------------------------------------------------------
void MagnetType::write_IsFrozen(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MagnetType::write_IsFrozen(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve number of write values
	int	w_length = attr.get_write_value_length();

	//	Retrieve pointer on write values (Do not delete !)
	const Tango::DevBoolean	*w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(MagnetType::write_IsFrozen) ENABLED START -----*/

  // Check number of received data
  if ((size_t)w_length != nb_magnet)
    Tango::Except::throw_exception(
            "WrongArg",
            "Wrong number of received data (" + std::to_string(nb_magnet) + " expected but got " + std::to_string(w_length) + ")",
            "MagnetType::write_IsFrozen"
    );

  vector<Tango::DevBoolean> set;
  set.assign(w_val,w_val+w_length);
  write_attribute<Tango::DevBoolean>("Frozen",(Tango::DevBoolean *)w_val,w_length);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::write_IsFrozen
}

//--------------------------------------------------------
/**
 *	Method     : MagnetType::add_dynamic_attributes()
 *	Description: Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void MagnetType::add_dynamic_attributes()
{
	/*----- PROTECTED REGION ID(MagnetType::add_dynamic_attributes) ENABLED START -----*/

	//	Add your own code to create and add dynamic attributes if any

	/*----- PROTECTED REGION END -----*/	//	MagnetType::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Command State related method
 *	Description: This command gets the device state (stored in its device_state data member) and returns it to the caller.
 *
 *	@returns Device state
 */
//--------------------------------------------------------
Tango::DevState MagnetType::dev_state()
{
	DEBUG_STREAM << "MagnetType::State()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::dev_state) ENABLED START -----*/

  Tango::DevState	argout; // replace by your own algorithm

  read_MagnetStates();

  try
  {
    vector<size_t> unk_devs;
    vector<size_t> fault_devs;
    vector<size_t> alarm_devs;
    vector<size_t> off_devs;
    vector<size_t> disa_devs;

    for (size_t loop = 0;loop < nb_magnet;loop++)
    {
      Tango::DevState sta = attr_MagnetStates_read[loop];

      switch (sta)
      {
        case Tango::UNKNOWN:
          unk_devs.push_back(loop);
          break;

        case Tango::FAULT:
          fault_devs.push_back(loop);
          break;

        case Tango::ALARM:
          alarm_devs.push_back(loop);
          break;

        case Tango::OFF:
          off_devs.push_back(loop);
          break;

        case Tango::DISABLE:
          disa_devs.push_back(loop);
          break;

        default:
          break;
      }
    }

    if (unk_devs.empty() == false)
    {
      argout = Tango::UNKNOWN;
      status_message(unk_devs,Tango::UNKNOWN);
    }
    else if (fault_devs.empty() == false)
    {
      argout = Tango::FAULT;
      status_message(fault_devs,Tango::FAULT);
    }
    else if (alarm_devs.empty() == false)
    {
      argout = Tango::ALARM;
      status_message(alarm_devs,Tango::ALARM);
    }
    else if (off_devs.empty() == false)
    {
      argout = Tango::OFF;
      status_message(off_devs,Tango::OFF);
    }
    else if (disa_devs.empty() == false)
    {
      argout = Tango::DISABLE;
      status_message(disa_devs,Tango::DISABLE);
    }
    else
    {
      argout = Tango::ON;
      set_status(Tango::StatusNotSet);
    }
  }
  catch (Tango::DevFailed &e)
  {
    argout = Tango::UNKNOWN;
    stringstream ss;
    ss << "Device in UNKNOWN state due to exception when getting low-level devices state";
    ss << "\nError: " << e.errors[0].desc.in();
    set_status(ss.str());
  }

	/*----- PROTECTED REGION END -----*/	//	MagnetType::dev_state
	set_state(argout);    // Give the state to Tango.
	if (argout!=Tango::ALARM)
		Tango::DeviceImpl::dev_state();
	return get_state();  // Return it after Tango management.
}
//--------------------------------------------------------
/**
 *	Command Reset related method
 *	Description: Reset device
 *
 */
//--------------------------------------------------------
void MagnetType::reset()
{
	DEBUG_STREAM << "MagnetType::Reset()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::reset) ENABLED START -----*/

  Tango::GroupCmdReplyList gcrl = grp->command_inout("Reset");

  string err_mess;
  for (size_t loop = 0;loop < gcrl.size();loop++)
  {
    if (gcrl[loop].has_failed())
    {
      const Tango::DevErrorList& del = gcrl[loop].get_err_stack();
      string dev_name = grp->get_device(loop + 1)->name();
      string tmp;
      tmp = "Error for device " + dev_name + ": " + del[0].desc.in();
      if (!err_mess.empty())
        err_mess.append("\n");
      err_mess.append(tmp);
    }
  }

  if (!err_mess.empty()) {
    Tango::Except::throw_exception("CommandError",err_mess,
                                   "MagnetType::reset()");
  }

	/*----- PROTECTED REGION END -----*/	//	MagnetType::reset
}
//--------------------------------------------------------
/**
 *	Command ResetResonanceStrengths related method
 *
 *
 */
//--------------------------------------------------------
void MagnetType::reset_resonance_strengths()
{
	DEBUG_STREAM << "MagnetType::ResetResonanceStrengths()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::reset_resonance_strengths) ENABLED START -----*/
  Tango::GroupCmdReplyList gcrl = grp->command_inout("ResetResonanceStrength");

  string err_mess;
  for (size_t loop = 0;loop < gcrl.size();loop++)
  {
    if (gcrl[loop].has_failed())
    {
      const Tango::DevErrorList& del = gcrl[loop].get_err_stack();
      string dev_name = grp->get_device(loop + 1)->name();
      string tmp;
      tmp = "Error for device " + dev_name + ": " + del[0].desc.in();
      if (!err_mess.empty())
        err_mess.append("\n");
      err_mess.append(tmp);
    }
  }

  if (!err_mess.empty()) {
    Tango::Except::throw_exception("CommandError",err_mess,
                                   "MagnetType::ResetResonanceStrength()");
  }

	/*----- PROTECTED REGION END -----*/	//	MagnetType::reset_resonance_strengths
}
//--------------------------------------------------------
/**
 *	Command SetpointCheck related method
 *	Description: Check for each family member, if the magnet is able to applied
 *               the new strengths. Returned the list of device(s) which are not able
 *               to applied the required strength. This list is empty if all
 *               magnets are able to generate the required strength
 *
 *	@param argin List of new strength for each magnet family member
 *	@returns List of magnet device name(s) which refuses their new strength
 */
//--------------------------------------------------------
Tango::DevVarStringArray *MagnetType::setpoint_check(const Tango::DevVarDoubleArray *argin)
{
	Tango::DevVarStringArray *argout;
	DEBUG_STREAM << "MagnetType::SetpointCheck()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(MagnetType::setpoint_check) ENABLED START -----*/

  // Check number of received data
  if (argin->length() != nb_magnet)
    Tango::Except::throw_exception(
            "WrongArg",
            "Wrong number of elt in file (" + std::to_string(nb_magnet) + " expected, but got " + std::to_string(argin->length()) + ")",
            "MagnetType::setpoint_check()"
    );

  vector<double> set;
  set.assign((*argin).get_buffer(),(*argin).get_buffer()+argin->length());
  Tango::GroupCmdReplyList gcrl = grp->command_inout("SetpointCheck",set);
  vector<size_t> refused_setpoint;

  string err_mess;
  for (size_t loop = 0;loop < gcrl.size();loop++)
  {
    if (gcrl[loop].has_failed()) {
      const Tango::DevErrorList& del = gcrl[loop].get_err_stack();
      string dev_name = grp->get_device(loop + 1)->name();
      string tmp;
      tmp = "Error for device " + dev_name + ": " + del[0].desc.in();
      if (!err_mess.empty())
        err_mess.append("\n");
      err_mess.append(tmp);
    } else {
      Tango::DevBoolean b;
      gcrl[loop] >> b;
      if(!b) refused_setpoint.emplace_back(loop);
    }
  }

  if (!err_mess.empty()) {
    Tango::Except::throw_exception("CommandError",err_mess,
                                   "MagnetType::SetpointCheck()");
  }

  argout = new Tango::DevVarStringArray();
  argout->length(refused_setpoint.size());
  size_t ctr = 0;
  for (const auto &elem : refused_setpoint)
    (*argout)[ctr++] = Tango::string_dup(attr_MagnetNames_read[elem]);

	/*----- PROTECTED REGION END -----*/	//	MagnetType::setpoint_check
	return argout;
}
//--------------------------------------------------------
/**
 *	Method     : MagnetType::add_dynamic_commands()
 *	Description: Create the dynamic commands if any
 *                for specified device.
 */
//--------------------------------------------------------
void MagnetType::add_dynamic_commands()
{
	/*----- PROTECTED REGION ID(MagnetType::add_dynamic_commands) ENABLED START -----*/

	//	Add your own code to create and add dynamic commands if any

	/*----- PROTECTED REGION END -----*/	//	MagnetType::add_dynamic_commands
}

/*----- PROTECTED REGION ID(MagnetType::namespace_ending) ENABLED START -----*/

//+------------------------------------------------------------------------------------------------------------------
//
// method :
//      MagnetType::local_read_MagnetStates
//
// description :
//      Read family devices MagnetStates attribute. This code is in a separate method in order to solve
//      a memory leak.
//
//-------------------------------------------------------------------------------------------------------------------

void MagnetType::read_MagnetStates()
{

  Tango::GroupAttrReplyList arl = grp->read_attribute("State");

  for (size_t i = 0;i < arl.size();i++) {
    if (arl[i].has_failed()) {
      attr_MagnetStates_read[i] = Tango::UNKNOWN;
      cout << attr_MagnetNames_read[i] << ":" << arl[i].get_err_stack()[0].desc << endl;
    } else {
      if (arl[i].get_data().get_quality() == Tango::ATTR_INVALID) {
        attr_MagnetStates_read[i] = Tango::UNKNOWN;
        cout << attr_MagnetNames_read[i] << ": INVALID" << endl;
      } else {
        arl[i] >> attr_MagnetStates_read[i];
      }
    }
  }

}

template<typename inType>
void MagnetType::write_attribute(const string& attName,inType *values,int length) {

  if(grp->get_size() != length) {
    Tango::Except::throw_exception("CantWriteAttribute",
                                   "Unexpected number of values, got " + to_string(length) + ", " +
                                   to_string(grp->get_size()) + " expected",
                                   "MagnetType::write_"+attName+"()");
  }

  vector<Tango::DeviceAttribute> das;
  for(int i=0;i<length;i++) {
    Tango::DeviceAttribute da(attName.c_str(), values[i]);
    das.push_back(da);
  }

  Tango::GroupReplyList grl = grp->write_attribute(das);

  string err_mess;
  for (size_t loop = 0;loop < grl.size();loop++)
  {
    if (grl[loop].has_failed())
    {
      const Tango::DevErrorList& del = grl[loop].get_err_stack();
      string dev_name = grp->get_device(loop + 1)->name();
      string tmp;
      tmp = "Error for device " + dev_name + ": " + del[0].desc.in();
      if (!err_mess.empty())
        err_mess.append("\n");
      err_mess.append(tmp);
    }
  }

  if (!err_mess.empty())
  {
    Tango::Except::throw_exception("CantWriteAttribute",err_mess,
                                   "MagnetType::write_"+attName+"()");
  }

}

//+------------------------------------------------------------------------------------------------------------------
//
// method :
//      MagnetType::compute_MSMM
// description :
//      Compute mean, standard deviation, minumum and maximum of the values in the vector given as first parameter.
//      std is computed as sqrt(mean(square) - square(mean))
//
// argument :
//      in:
//          - _data : The input data
//      out:
//          - _msmm : Output vector with mean, std, min and max
//
//-------------------------------------------------------------------------------------------------------------------

void MagnetType::compute_MSMM(double *_data,size_t _nb_data,vector<double> &_msmm)
{
    double sum, sum2, min, max;
    sum = sum2 = 0.0;

    size_t nb_nan = 0;

    min = _data[0];
    max = _data[0];

    for (size_t loop = 0;loop < _nb_data;loop++)
    {
        double data = _data[loop];
        if (isnan(data) == false)
        {
            if (data > max)
                max = data;
            if (data < min)
                min = data;
            sum += data;
            sum2 += data * data;
        }
        else
            nb_nan++;
    }

    double nb_data = (double)(_nb_data - nb_nan);
    double aver = sum / nb_data;
    double std = sqrt((sum2 / nb_data) - (aver * aver));

    _msmm[0] = aver;
    _msmm[1] = std;
    _msmm[2] = min;
    _msmm[3] = max;
}

//+------------------------------------------------------------------------------------------------------------------
//
// method :
//      MagnetType::set_att_unit
// description :
//      Set attribute unit as a function of the magnet type
//
// argument :
//      in:
//          - _att_name : The attribute name
//
//-------------------------------------------------------------------------------------------------------------------

void MagnetType::set_att_unit(const string &_att_name)
{
    Tango::Attribute &att = get_device_attr()->get_attr_by_name(_att_name.c_str());
    Tango::MultiAttrProp<Tango::DevDouble> att_prop;

    att.get_properties(att_prop);

    if (mag_type == MagType::QUAD)
    {
        att_prop.unit = "m-1";
    }
    else if (mag_type == MagType::OCTU)
    {
        att_prop.unit = "m-3";
    }
    else
    {
        att_prop.unit = "m-2";
    }

    att.set_properties(att_prop);
}

/*----- PROTECTED REGION END -----*/	//	MagnetType::namespace_ending
} //	namespace
