/*----- PROTECTED REGION ID(MagnetType.h) ENABLED START -----*/
//=============================================================================
//
// file :        MagnetType.h
//
// description : Include file for the MagnetType class
//
// project :     MagnetType
//
// This file is part of Tango device class.
//
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
//
// $Author:  $
// Copyright (C): 2004,2005,2006,2007,2008,2009,2009,2010,2011,2012,2013,2014,2015
//                European Synchrotron Radiation Facility
//                BP 220, Grenoble 38043
//                France
//
// $Revision:  $
// $Date:  $
//
// $HeadURL:  $
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#ifndef MagnetType_H
#define MagnetType_H

#include <tango.h>
#include <chrono>


/*----- PROTECTED REGION END -----*/	//	MagnetType.h

#ifdef TANGO_LOG
	// cppTango after c934adea (Merge branch 'remove-cout-definition' into 'main', 2022-05-23)
	// nothing to do
#else
	// cppTango 9.3-backports and older
	#define TANGO_LOG       cout
	#define TANGO_LOG_INFO  cout2
	#define TANGO_LOG_DEBUG cout3
#endif // TANGO_LOG

/**
 *  MagnetType class description:
 *    This class is for the ESRF EBS storage ring simulator.
 *    It is the interface for a whole set of magnets from the same type
 *    (all quadrupoles, all sextupoles,...). This class is supposed to
 *    be used by control loop software
 */


namespace MagnetType_ns
{
/*----- PROTECTED REGION ID(MagnetType::Additional Class Declarations) ENABLED START -----*/

const size_t QUAD_NB =              8;
const size_t SEXTU_NB =             192;
const size_t OCTU_NB =              64;
const size_t DQ_NB =                0;
const size_t MAX_MAGNET_NB =        610;

const size_t NB_QUAD_CELL =         19;

/*----- PROTECTED REGION END -----*/	//	MagnetType::Additional Class Declarations

class MagnetType : public TANGO_BASE_CLASS
{

/*----- PROTECTED REGION ID(MagnetType::Data Members) ENABLED START -----*/

    enum class _MagType
    {
        QUAD = 0,
        SEXTU,
        OCTU,
    };
    typedef _MagType MagType;

private:
    Tango::Group            *grp;
    MagType                 mag_type;
    size_t                  nb_magnet;
    Tango::DevDouble        *attr_CorrectionStrengths_write;
    Tango::DevDouble        *attr_ResonanceStrengths_write;
    Tango::DevDouble        *attr_Strengths_write;
    Tango::DevDouble        *attr_TotalCorrectionStrengths_write;
    Tango::DevBoolean       *attr_IsFrozen_write;
    string                  design_file_name;

    vector<double>          corr_msmm;
    vector<double>          reso_msmm;
    vector<double>          str_msmm;
    vector<double>          total_msmm;

    vector<double>          design_strength;

/*----- PROTECTED REGION END -----*/	//	MagnetType::Data Members

//	Device property data members
public:
	//	MagnetType:	Name of  the magnet type
	std::string	magnetType;
	//	MagnetNames:	List of magnet devices
	std::vector<std::string>	magnetNames;

	bool	mandatoryNotDefined;

//	Attribute data members
public:
	Tango::DevString	*attr_DesignStrengthFile_read;
	Tango::DevLong	*attr_CorrectionStrengthsSteps_read;
	Tango::DevDouble	*attr_Strengths_read;
	Tango::DevDouble	*attr_StrengthsMSMM_read;
	Tango::DevDouble	*attr_StrengthsReadPoint_read;
	Tango::DevDouble	*attr_TotalCorrectionStrengths_read;
	Tango::DevDouble	*attr_TotalCorrectionStrengthsMSMM_read;
	Tango::DevDouble	*attr_CorrectionStrengths_read;
	Tango::DevDouble	*attr_CorrectionStrengthsMSMM_read;
	Tango::DevDouble	*attr_ResonanceStrengths_read;
	Tango::DevDouble	*attr_ResonanceStrengthsMSMM_read;
	Tango::DevString	*attr_MagnetNames_read;
	Tango::DevState	*attr_MagnetStates_read;
	Tango::DevBoolean	*attr_RequireCycling_read;
	Tango::DevString	*attr_FrozenMagnets_read;
	Tango::DevBoolean	*attr_IsFrozen_read;

//	Constructors and destructors
public:
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device Name
	 */
	MagnetType(Tango::DeviceClass *cl,std::string &s);
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device Name
	 */
	MagnetType(Tango::DeviceClass *cl,const char *s);
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device name
	 *	@param d	Device description.
	 */
	MagnetType(Tango::DeviceClass *cl,const char *s,const char *d);
	/**
	 * The device object destructor.
	 */
	~MagnetType();


//	Miscellaneous methods
public:
	/*
	 *	will be called at device destruction or at init command.
	 */
	void delete_device();
	/*
	 *	Initialize the device
	 */
	virtual void init_device();
	/*
	 *	Read the device properties from database
	 */
	void get_device_property();
	/*
	 *	Always executed method before execution command method.
	 */
	virtual void always_executed_hook();

	/*
	 *	Check if mandatory property has been set
	 */
	 void check_mandatory_property(Tango::DbDatum &class_prop, Tango::DbDatum &dev_prop);

//	Attribute methods
public:
	//--------------------------------------------------------
	/*
	 *	Method     : MagnetType::read_attr_hardware()
	 *	Description: Hardware acquisition for attributes.
	 */
	//--------------------------------------------------------
	virtual void read_attr_hardware(std::vector<long> &attr_list);
	//--------------------------------------------------------
	/*
	 *	Method     : MagnetType::write_attr_hardware()
	 *	Description: Hardware writing for attributes.
	 */
	//--------------------------------------------------------
	virtual void write_attr_hardware(std::vector<long> &attr_list);

/**
 *	Attribute DesignStrengthFile related methods
 *	Description: Design strength file name
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
	virtual void read_DesignStrengthFile(Tango::Attribute &attr);
	virtual void write_DesignStrengthFile(Tango::WAttribute &attr);
	virtual bool is_DesignStrengthFile_allowed(Tango::AttReqType type);
/**
 *	Attribute CorrectionStrengthsSteps related methods
 *	Description: Number of steps in which a new Correction Strenths must be applied
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
	virtual void read_CorrectionStrengthsSteps(Tango::Attribute &attr);
	virtual void write_CorrectionStrengthsSteps(Tango::WAttribute &attr);
	virtual bool is_CorrectionStrengthsSteps_allowed(Tango::AttReqType type);
/**
 *	Attribute Strengths related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 610
 */
	virtual void read_Strengths(Tango::Attribute &attr);
	virtual bool is_Strengths_allowed(Tango::AttReqType type);
/**
 *	Attribute StrengthsMSMM related methods
 *	Description: Strengths spectrum attribute Mean, Std, Min and Max
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4
 */
	virtual void read_StrengthsMSMM(Tango::Attribute &attr);
	virtual bool is_StrengthsMSMM_allowed(Tango::AttReqType type);
/**
 *	Attribute StrengthsReadPoint related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 610
 */
	virtual void read_StrengthsReadPoint(Tango::Attribute &attr);
	virtual bool is_StrengthsReadPoint_allowed(Tango::AttReqType type);
/**
 *	Attribute TotalCorrectionStrengths related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 610
 */
	virtual void read_TotalCorrectionStrengths(Tango::Attribute &attr);
	virtual bool is_TotalCorrectionStrengths_allowed(Tango::AttReqType type);
/**
 *	Attribute TotalCorrectionStrengthsMSMM related methods
 *	Description: TotalCorrectionStrengths spectrum attribute Mean, Std, Min and Max
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4
 */
	virtual void read_TotalCorrectionStrengthsMSMM(Tango::Attribute &attr);
	virtual bool is_TotalCorrectionStrengthsMSMM_allowed(Tango::AttReqType type);
/**
 *	Attribute CorrectionStrengths related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 610
 */
	virtual void read_CorrectionStrengths(Tango::Attribute &attr);
	virtual void write_CorrectionStrengths(Tango::WAttribute &attr);
	virtual bool is_CorrectionStrengths_allowed(Tango::AttReqType type);
/**
 *	Attribute CorrectionStrengthsMSMM related methods
 *	Description: CorrectionStrengths spectrum attribute Mean, Std, Min and Max
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4
 */
	virtual void read_CorrectionStrengthsMSMM(Tango::Attribute &attr);
	virtual bool is_CorrectionStrengthsMSMM_allowed(Tango::AttReqType type);
/**
 *	Attribute ResonanceStrengths related methods
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 610
 */
	virtual void read_ResonanceStrengths(Tango::Attribute &attr);
	virtual void write_ResonanceStrengths(Tango::WAttribute &attr);
	virtual bool is_ResonanceStrengths_allowed(Tango::AttReqType type);
/**
 *	Attribute ResonanceStrengthsMSMM related methods
 *	Description: ResonanceStrengths spectrum attribute Mean, Std, Min and Max
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4
 */
	virtual void read_ResonanceStrengthsMSMM(Tango::Attribute &attr);
	virtual bool is_ResonanceStrengthsMSMM_allowed(Tango::AttReqType type);
/**
 *	Attribute MagnetNames related methods
 *	Description: Level 1 magnet device names
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Spectrum max = 610
 */
	virtual void read_MagnetNames(Tango::Attribute &attr);
	virtual bool is_MagnetNames_allowed(Tango::AttReqType type);
/**
 *	Attribute MagnetStates related methods
 *
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Spectrum max = 620
 */
	virtual void read_MagnetStates(Tango::Attribute &attr);
	virtual bool is_MagnetStates_allowed(Tango::AttReqType type);
/**
 *	Attribute RequireCycling related methods
 *	Description: Boolean set for each magnets which requires cycling (in beam propagation order)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Spectrum max = 620
 */
	virtual void read_RequireCycling(Tango::Attribute &attr);
	virtual bool is_RequireCycling_allowed(Tango::AttReqType type);
/**
 *	Attribute FrozenMagnets related methods
 *	Description: List of disabled magnets
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Spectrum max = 620
 */
	virtual void read_FrozenMagnets(Tango::Attribute &attr);
	virtual bool is_FrozenMagnets_allowed(Tango::AttReqType type);
/**
 *	Attribute IsFrozen related methods
 *	Description: Boolean set for each disabled magnets (in beam propagation order)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Spectrum max = 620
 */
	virtual void read_IsFrozen(Tango::Attribute &attr);
	virtual void write_IsFrozen(Tango::WAttribute &attr);
	virtual bool is_IsFrozen_allowed(Tango::AttReqType type);


	//--------------------------------------------------------
	/**
	 *	Method     : MagnetType::add_dynamic_attributes()
	 *	Description: Add dynamic attributes if any.
	 */
	//--------------------------------------------------------
	void add_dynamic_attributes();




//	Command related methods
public:
	/**
	 *	Command State related method
	 *	Description: This command gets the device state (stored in its device_state data member) and returns it to the caller.
	 *
	 *	@returns Device state
	 */
	virtual Tango::DevState dev_state();
	/**
	 *	Command Reset related method
	 *	Description: Reset device
	 *
	 */
	virtual void reset();
	virtual bool is_Reset_allowed(const CORBA::Any &any);
	/**
	 *	Command ResetResonanceStrengths related method
	 *
	 *
	 */
	virtual void reset_resonance_strengths();
	virtual bool is_ResetResonanceStrengths_allowed(const CORBA::Any &any);
	/**
	 *	Command SetpointCheck related method
	 *	Description: Check for each family member, if the magnet is able to applied
	 *               the new strengths. Returned the list of device(s) which are not able
	 *               to applied the required strength. This list is empty if all
	 *               magnets are able to generate the required strength
	 *
	 *	@param argin List of new strength for each magnet family member
	 *	@returns List of magnet device name(s) which refuses their new strength
	 */
	virtual Tango::DevVarStringArray *setpoint_check(const Tango::DevVarDoubleArray *argin);
	virtual bool is_SetpointCheck_allowed(const CORBA::Any &any);


	//--------------------------------------------------------
	/**
	 *	Method     : MagnetType::add_dynamic_commands()
	 *	Description: Add dynamic commands if any.
	 */
	//--------------------------------------------------------
	void add_dynamic_commands();

/*----- PROTECTED REGION ID(MagnetType::Additional Method prototypes) ENABLED START -----*/

private:

    template<typename inType>
    void write_attribute(const string& attName,inType *values,int length);
    int devname_2_ind(const string &);
    void compute_MSMM(double *,size_t,vector<double> &);
    void set_att_unit(const string &);
    void status_message(vector<size_t> &,Tango::DevState);
    void read_MagnetStates();

/*----- PROTECTED REGION END -----*/	//	MagnetType::Additional Method prototypes
};

/*----- PROTECTED REGION ID(MagnetType::Additional Classes Definitions) ENABLED START -----*/

inline void MagnetType::status_message(vector<size_t> &_dev_inds,Tango::DevState _sta)
{
    if (_dev_inds.size() == nb_magnet)
    {
        stringstream ss;
        if (_sta == Tango::DISABLE)
            ss << "All " << nb_magnet << " are FROZEN";
        else
            ss << "All " << nb_magnet << " devices are in " << Tango::DevStateName[_sta] << " state";
        set_status(ss.str());
    }
    else
    {
        stringstream ss;
        ss << "Device(s) ";
        for (const auto& elem : _dev_inds)
            ss << attr_MagnetNames_read[elem] << ", ";

        if (_sta == Tango::DISABLE)
            ss << " are FROZEN";
        else
            ss << "are in " << Tango::DevStateName[_sta] << " state";
        set_status(ss.str());
    }
}

/*----- PROTECTED REGION END -----*/	//	MagnetType::Additional Classes Definitions

}	//	End of namespace

#endif   //	MagnetType_H
