/*----- PROTECTED REGION ID(MagnetTypeClass.cpp) ENABLED START -----*/
//=============================================================================
//
// file :        MagnetTypeClass.cpp
//
// description : C++ source for the MagnetTypeClass.
//               A singleton class derived from DeviceClass.
//               It implements the command and attribute list
//               and all properties and methods required
//               by the MagnetType once per process.
//
// project :     MagnetTypePS
//
// This file is part of Tango device class.
// 
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
// 
//
// Copyright (C): 2004,2005,2006,2007,2008,2009,2009,2010,2011,2012,2013,2014,2015
//                European Synchrotron Radiation Facility
//                BP 220, Grenoble 38043
//                France
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include <MagnetTypeClass.h>

/*----- PROTECTED REGION END -----*/	//	MagnetTypeClass.cpp

//-------------------------------------------------------------------
/**
 *	Create MagnetTypeClass singleton and
 *	return it in a C function for Python usage
 */
//-------------------------------------------------------------------
extern "C" {
#ifdef _TG_WINDOWS_

__declspec(dllexport)

#endif

	Tango::DeviceClass *_create_MagnetType_class(const char *name) {
		return MagnetType_ns::MagnetTypeClass::init(name);
	}
}

namespace MagnetType_ns
{
//===================================================================
//	Initialize pointer for singleton pattern
//===================================================================
MagnetTypeClass *MagnetTypeClass::_instance = NULL;

//===================================================================
//	Class constants
//===================================================================
constexpr long StrengthsAttrib::X_DATA_SIZE;
constexpr long StrengthsMSMMAttrib::X_DATA_SIZE;
constexpr long StrengthsReadPointAttrib::X_DATA_SIZE;
constexpr long TotalCorrectionStrengthsAttrib::X_DATA_SIZE;
constexpr long TotalCorrectionStrengthsMSMMAttrib::X_DATA_SIZE;
constexpr long CorrectionStrengthsAttrib::X_DATA_SIZE;
constexpr long CorrectionStrengthsMSMMAttrib::X_DATA_SIZE;
constexpr long ResonanceStrengthsAttrib::X_DATA_SIZE;
constexpr long ResonanceStrengthsMSMMAttrib::X_DATA_SIZE;
constexpr long MagnetNamesAttrib::X_DATA_SIZE;
constexpr long MagnetStatesAttrib::X_DATA_SIZE;
constexpr long RequireCyclingAttrib::X_DATA_SIZE;
constexpr long FrozenMagnetsAttrib::X_DATA_SIZE;
constexpr long IsFrozenAttrib::X_DATA_SIZE;
//--------------------------------------------------------
/**
 * method : 		MagnetTypeClass::MagnetTypeClass(std::string &s)
 * description : 	constructor for the MagnetTypeClass
 *
 * @param s	The class name
 */
//--------------------------------------------------------
MagnetTypeClass::MagnetTypeClass(std::string &s):Tango::DeviceClass(s)
{
	TANGO_LOG_INFO << "Entering MagnetTypeClass constructor" << std::endl;
	set_default_property();
	write_class_property();

	/*----- PROTECTED REGION ID(MagnetTypeClass::constructor) ENABLED START -----*/
	
	/*----- PROTECTED REGION END -----*/	//	MagnetTypeClass::constructor

	TANGO_LOG_INFO << "Leaving MagnetTypeClass constructor" << std::endl;
}

//--------------------------------------------------------
/**
 * method : 		MagnetTypeClass::~MagnetTypeClass()
 * description : 	destructor for the MagnetTypeClass
 */
//--------------------------------------------------------
MagnetTypeClass::~MagnetTypeClass()
{
	/*----- PROTECTED REGION ID(MagnetTypeClass::destructor) ENABLED START -----*/
	
	/*----- PROTECTED REGION END -----*/	//	MagnetTypeClass::destructor

	_instance = NULL;
}


//--------------------------------------------------------
/**
 * method : 		MagnetTypeClass::init
 * description : 	Create the object if not already done.
 *                  Otherwise, just return a pointer to the object
 *
 * @param	name	The class name
 */
//--------------------------------------------------------
MagnetTypeClass *MagnetTypeClass::init(const char *name)
{
	if (_instance == NULL)
	{
		try
		{
			std::string s(name);
			_instance = new MagnetTypeClass(s);
		}
		catch (std::bad_alloc &)
		{
			throw;
		}
	}
	return _instance;
}

//--------------------------------------------------------
/**
 * method : 		MagnetTypeClass::instance
 * description : 	Check if object already created,
 *                  and return a pointer to the object
 */
//--------------------------------------------------------
MagnetTypeClass *MagnetTypeClass::instance()
{
	if (_instance == NULL)
	{
		std::cerr << "Class is not initialized !!" << std::endl;
		exit(-1);
	}
	return _instance;
}



//===================================================================
//	Command execution method calls
//===================================================================
//--------------------------------------------------------
/**
 * method : 		ResetClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *ResetClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "ResetClass::execute(): arrived" << std::endl;
	((static_cast<MagnetType *>(device))->reset());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		ResetResonanceStrengthsClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *ResetResonanceStrengthsClass::execute(Tango::DeviceImpl *device, TANGO_UNUSED(const CORBA::Any &in_any))
{
	TANGO_LOG_INFO << "ResetResonanceStrengthsClass::execute(): arrived" << std::endl;
	((static_cast<MagnetType *>(device))->reset_resonance_strengths());
	return new CORBA::Any();
}

//--------------------------------------------------------
/**
 * method : 		SetpointCheckClass::execute()
 * description : 	method to trigger the execution of the command.
 *
 * @param	device	The device on which the command must be executed
 * @param	in_any	The command input data
 *
 *	returns The command output data (packed in the Any object)
 */
//--------------------------------------------------------
CORBA::Any *SetpointCheckClass::execute(Tango::DeviceImpl *device, const CORBA::Any &in_any)
{
	TANGO_LOG_INFO << "SetpointCheckClass::execute(): arrived" << std::endl;
	const Tango::DevVarDoubleArray *argin;
	extract(in_any, argin);
	return insert((static_cast<MagnetType *>(device))->setpoint_check(argin));
}


//===================================================================
//	Properties management
//===================================================================
//--------------------------------------------------------
/**
 *	Method     : MagnetTypeClass::get_class_property()
 *	Description: Get the class property for specified name.
 */
//--------------------------------------------------------
Tango::DbDatum MagnetTypeClass::get_class_property(std::string &prop_name)
{
	for (unsigned int i=0 ; i<cl_prop.size() ; i++)
		if (cl_prop[i].name == prop_name)
			return cl_prop[i];
	//	if not found, returns  an empty DbDatum
	return Tango::DbDatum(prop_name);
}

//--------------------------------------------------------
/**
 *	Method     : MagnetTypeClass::get_default_device_property()
 *	Description: Return the default value for device property.
 */
//--------------------------------------------------------
Tango::DbDatum MagnetTypeClass::get_default_device_property(std::string &prop_name)
{
	for (unsigned int i=0 ; i<dev_def_prop.size() ; i++)
		if (dev_def_prop[i].name == prop_name)
			return dev_def_prop[i];
	//	if not found, return  an empty DbDatum
	return Tango::DbDatum(prop_name);
}

//--------------------------------------------------------
/**
 *	Method     : MagnetTypeClass::get_default_class_property()
 *	Description: Return the default value for class property.
 */
//--------------------------------------------------------
Tango::DbDatum MagnetTypeClass::get_default_class_property(std::string &prop_name)
{
	for (unsigned int i=0 ; i<cl_def_prop.size() ; i++)
		if (cl_def_prop[i].name == prop_name)
			return cl_def_prop[i];
	//	if not found, return  an empty DbDatum
	return Tango::DbDatum(prop_name);
}


//--------------------------------------------------------
/**
 *	Method     : MagnetTypeClass::set_default_property()
 *	Description: Set default property (class and device) for wizard.
 *                For each property, add to wizard property name and description.
 *                If default value has been set, add it to wizard property and
 *                store it in a DbDatum.
 */
//--------------------------------------------------------
void MagnetTypeClass::set_default_property()
{
	std::string	prop_name;
	std::string	prop_desc;
	std::string	prop_def;
	std::vector<std::string>	vect_data;

	//	Set Default Class Properties

	//	Set Default device Properties
	prop_name = "MagnetType";
	prop_desc = "Name of  the magnet type";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
	prop_name = "MagnetNames";
	prop_desc = "List of magnet devices";
	prop_def  = "";
	vect_data.clear();
	if (prop_def.length()>0)
	{
		Tango::DbDatum	data(prop_name);
		data << vect_data ;
		dev_def_prop.push_back(data);
		add_wiz_dev_prop(prop_name, prop_desc,  prop_def);
	}
	else
		add_wiz_dev_prop(prop_name, prop_desc);
}

//--------------------------------------------------------
/**
 *	Method     : MagnetTypeClass::write_class_property()
 *	Description: Set class description fields as property in database
 */
//--------------------------------------------------------
void MagnetTypeClass::write_class_property()
{
	//	First time, check if database used
	if (Tango::Util::_UseDb == false)
		return;

	Tango::DbData	data;
	std::string	classname = get_name();
	std::string	header;

	//	Put title
	Tango::DbDatum	title("ProjectTitle");
	std::string	str_title("MagnetType");
	title << str_title;
	data.push_back(title);

	//	Put Description
	Tango::DbDatum	description("Description");
	std::vector<std::string>	str_desc;
	str_desc.push_back("This class is for the ESRF EBS storage ring simulator.");
	str_desc.push_back("It is the interface for a whole set of magnets from the same type");
	str_desc.push_back("(all quadrupoles, all sextupoles,...). This class is supposed to");
	str_desc.push_back("be used by control loop software");
	description << str_desc;
	data.push_back(description);

	//  Put inheritance
	Tango::DbDatum	inher_datum("InheritedFrom");
	std::vector<std::string> inheritance;
	inheritance.push_back("TANGO_BASE_CLASS");
	inher_datum << inheritance;
	data.push_back(inher_datum);

	//	Call database and and values
	get_db_class()->put_property(data);
}

//===================================================================
//	Factory methods
//===================================================================

//--------------------------------------------------------
/**
 *	Method     : MagnetTypeClass::device_factory()
 *	Description: Create the device object(s)
 *                and store them in the device list
 */
//--------------------------------------------------------
void MagnetTypeClass::device_factory(const Tango::DevVarStringArray *devlist_ptr)
{
	/*----- PROTECTED REGION ID(MagnetTypeClass::device_factory_before) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	MagnetTypeClass::device_factory_before

	//	Create devices and add it into the device list
	for (unsigned long i=0 ; i<devlist_ptr->length() ; i++)
	{
		TANGO_LOG_DEBUG << "Device name : " << (*devlist_ptr)[i].in() << std::endl;
		device_list.push_back(new MagnetType(this, (*devlist_ptr)[i]));
	}

	//	Manage dynamic attributes if any
	erase_dynamic_attributes(devlist_ptr, get_class_attr()->get_attr_list());

	//	Export devices to the outside world
	for (unsigned long i=1 ; i<=devlist_ptr->length() ; i++)
	{
		//	Add dynamic attributes if any
		MagnetType *dev = static_cast<MagnetType *>(device_list[device_list.size()-i]);
		dev->add_dynamic_attributes();

		//	Check before if database used.
		if ((Tango::Util::_UseDb == true) && (Tango::Util::_FileDb == false))
			export_device(dev);
		else
			export_device(dev, dev->get_name().c_str());
	}

	/*----- PROTECTED REGION ID(MagnetTypeClass::device_factory_after) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	MagnetTypeClass::device_factory_after
}
//--------------------------------------------------------
/**
 *	Method     : MagnetTypeClass::attribute_factory()
 *	Description: Create the attribute object(s)
 *                and store them in the attribute list
 */
//--------------------------------------------------------
void MagnetTypeClass::attribute_factory(std::vector<Tango::Attr *> &att_list)
{
	/*----- PROTECTED REGION ID(MagnetTypeClass::attribute_factory_before) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	MagnetTypeClass::attribute_factory_before
	//	Attribute : DesignStrengthFile
	DesignStrengthFileAttrib	*designstrengthfile = new DesignStrengthFileAttrib();
	Tango::UserDefaultAttrProp	designstrengthfile_prop;
	designstrengthfile_prop.set_description("Design strength file name");
	designstrengthfile_prop.set_label("Design strength file");
	//	unit	not set for DesignStrengthFile
	//	standard_unit	not set for DesignStrengthFile
	//	display_unit	not set for DesignStrengthFile
	//	format	not set for DesignStrengthFile
	//	max_value	not set for DesignStrengthFile
	//	min_value	not set for DesignStrengthFile
	//	max_alarm	not set for DesignStrengthFile
	//	min_alarm	not set for DesignStrengthFile
	//	max_warning	not set for DesignStrengthFile
	//	min_warning	not set for DesignStrengthFile
	//	delta_t	not set for DesignStrengthFile
	//	delta_val	not set for DesignStrengthFile
	designstrengthfile->set_default_properties(designstrengthfile_prop);
	//	Not Polled
	designstrengthfile->set_disp_level(Tango::EXPERT);
	designstrengthfile->set_memorized();
	designstrengthfile->set_memorized_init(false);
	att_list.push_back(designstrengthfile);

	//	Attribute : CorrectionStrengthsSteps
	CorrectionStrengthsStepsAttrib	*correctionstrengthssteps = new CorrectionStrengthsStepsAttrib();
	Tango::UserDefaultAttrProp	correctionstrengthssteps_prop;
	correctionstrengthssteps_prop.set_description("Number of steps in which a new Correction Strenths must be applied");
	//	label	not set for CorrectionStrengthsSteps
	//	unit	not set for CorrectionStrengthsSteps
	//	standard_unit	not set for CorrectionStrengthsSteps
	//	display_unit	not set for CorrectionStrengthsSteps
	//	format	not set for CorrectionStrengthsSteps
	correctionstrengthssteps_prop.set_max_value("10");
	correctionstrengthssteps_prop.set_min_value("1");
	//	max_alarm	not set for CorrectionStrengthsSteps
	//	min_alarm	not set for CorrectionStrengthsSteps
	//	max_warning	not set for CorrectionStrengthsSteps
	//	min_warning	not set for CorrectionStrengthsSteps
	//	delta_t	not set for CorrectionStrengthsSteps
	//	delta_val	not set for CorrectionStrengthsSteps
	correctionstrengthssteps->set_default_properties(correctionstrengthssteps_prop);
	//	Not Polled
	correctionstrengthssteps->set_disp_level(Tango::EXPERT);
	correctionstrengthssteps->set_memorized();
	correctionstrengthssteps->set_memorized_init(true);
	att_list.push_back(correctionstrengthssteps);

	//	Attribute : Strengths
	StrengthsAttrib	*strengths = new StrengthsAttrib();
	Tango::UserDefaultAttrProp	strengths_prop;
	//	description	not set for Strengths
	//	label	not set for Strengths
	strengths_prop.set_unit("m-1");
	//	standard_unit	not set for Strengths
	//	display_unit	not set for Strengths
	strengths_prop.set_format("%4.3e");
	//	max_value	not set for Strengths
	//	min_value	not set for Strengths
	//	max_alarm	not set for Strengths
	//	min_alarm	not set for Strengths
	//	max_warning	not set for Strengths
	//	min_warning	not set for Strengths
	//	delta_t	not set for Strengths
	//	delta_val	not set for Strengths
	strengths->set_default_properties(strengths_prop);
	//	Not Polled
	strengths->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(strengths);

	//	Attribute : StrengthsMSMM
	StrengthsMSMMAttrib	*strengthsmsmm = new StrengthsMSMMAttrib();
	Tango::UserDefaultAttrProp	strengthsmsmm_prop;
	strengthsmsmm_prop.set_description("Strengths spectrum attribute Mean, Std, Min and Max");
	//	label	not set for StrengthsMSMM
	//	unit	not set for StrengthsMSMM
	//	standard_unit	not set for StrengthsMSMM
	//	display_unit	not set for StrengthsMSMM
	strengthsmsmm_prop.set_format("%4.3e");
	//	max_value	not set for StrengthsMSMM
	//	min_value	not set for StrengthsMSMM
	//	max_alarm	not set for StrengthsMSMM
	//	min_alarm	not set for StrengthsMSMM
	//	max_warning	not set for StrengthsMSMM
	//	min_warning	not set for StrengthsMSMM
	//	delta_t	not set for StrengthsMSMM
	//	delta_val	not set for StrengthsMSMM
	strengthsmsmm->set_default_properties(strengthsmsmm_prop);
	//	Not Polled
	strengthsmsmm->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(strengthsmsmm);

	//	Attribute : StrengthsReadPoint
	StrengthsReadPointAttrib	*strengthsreadpoint = new StrengthsReadPointAttrib();
	Tango::UserDefaultAttrProp	strengthsreadpoint_prop;
	//	description	not set for StrengthsReadPoint
	//	label	not set for StrengthsReadPoint
	strengthsreadpoint_prop.set_unit("m-1");
	//	standard_unit	not set for StrengthsReadPoint
	//	display_unit	not set for StrengthsReadPoint
	strengthsreadpoint_prop.set_format("%4.3e");
	//	max_value	not set for StrengthsReadPoint
	//	min_value	not set for StrengthsReadPoint
	//	max_alarm	not set for StrengthsReadPoint
	//	min_alarm	not set for StrengthsReadPoint
	//	max_warning	not set for StrengthsReadPoint
	//	min_warning	not set for StrengthsReadPoint
	//	delta_t	not set for StrengthsReadPoint
	//	delta_val	not set for StrengthsReadPoint
	strengthsreadpoint->set_default_properties(strengthsreadpoint_prop);
	//	Not Polled
	strengthsreadpoint->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(strengthsreadpoint);

	//	Attribute : TotalCorrectionStrengths
	TotalCorrectionStrengthsAttrib	*totalcorrectionstrengths = new TotalCorrectionStrengthsAttrib();
	Tango::UserDefaultAttrProp	totalcorrectionstrengths_prop;
	//	description	not set for TotalCorrectionStrengths
	//	label	not set for TotalCorrectionStrengths
	totalcorrectionstrengths_prop.set_unit("m-1");
	//	standard_unit	not set for TotalCorrectionStrengths
	//	display_unit	not set for TotalCorrectionStrengths
	totalcorrectionstrengths_prop.set_format("%4.3e");
	//	max_value	not set for TotalCorrectionStrengths
	//	min_value	not set for TotalCorrectionStrengths
	//	max_alarm	not set for TotalCorrectionStrengths
	//	min_alarm	not set for TotalCorrectionStrengths
	//	max_warning	not set for TotalCorrectionStrengths
	//	min_warning	not set for TotalCorrectionStrengths
	//	delta_t	not set for TotalCorrectionStrengths
	//	delta_val	not set for TotalCorrectionStrengths
	totalcorrectionstrengths->set_default_properties(totalcorrectionstrengths_prop);
	//	Not Polled
	totalcorrectionstrengths->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(totalcorrectionstrengths);

	//	Attribute : TotalCorrectionStrengthsMSMM
	TotalCorrectionStrengthsMSMMAttrib	*totalcorrectionstrengthsmsmm = new TotalCorrectionStrengthsMSMMAttrib();
	Tango::UserDefaultAttrProp	totalcorrectionstrengthsmsmm_prop;
	totalcorrectionstrengthsmsmm_prop.set_description("TotalCorrectionStrengths spectrum attribute Mean, Std, Min and Max");
	//	label	not set for TotalCorrectionStrengthsMSMM
	//	unit	not set for TotalCorrectionStrengthsMSMM
	//	standard_unit	not set for TotalCorrectionStrengthsMSMM
	//	display_unit	not set for TotalCorrectionStrengthsMSMM
	totalcorrectionstrengthsmsmm_prop.set_format("%4.3e");
	//	max_value	not set for TotalCorrectionStrengthsMSMM
	//	min_value	not set for TotalCorrectionStrengthsMSMM
	//	max_alarm	not set for TotalCorrectionStrengthsMSMM
	//	min_alarm	not set for TotalCorrectionStrengthsMSMM
	//	max_warning	not set for TotalCorrectionStrengthsMSMM
	//	min_warning	not set for TotalCorrectionStrengthsMSMM
	//	delta_t	not set for TotalCorrectionStrengthsMSMM
	//	delta_val	not set for TotalCorrectionStrengthsMSMM
	totalcorrectionstrengthsmsmm->set_default_properties(totalcorrectionstrengthsmsmm_prop);
	//	Not Polled
	totalcorrectionstrengthsmsmm->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(totalcorrectionstrengthsmsmm);

	//	Attribute : CorrectionStrengths
	CorrectionStrengthsAttrib	*correctionstrengths = new CorrectionStrengthsAttrib();
	Tango::UserDefaultAttrProp	correctionstrengths_prop;
	//	description	not set for CorrectionStrengths
	//	label	not set for CorrectionStrengths
	correctionstrengths_prop.set_unit("m-1");
	//	standard_unit	not set for CorrectionStrengths
	//	display_unit	not set for CorrectionStrengths
	correctionstrengths_prop.set_format("%4.3e");
	//	max_value	not set for CorrectionStrengths
	//	min_value	not set for CorrectionStrengths
	//	max_alarm	not set for CorrectionStrengths
	//	min_alarm	not set for CorrectionStrengths
	//	max_warning	not set for CorrectionStrengths
	//	min_warning	not set for CorrectionStrengths
	//	delta_t	not set for CorrectionStrengths
	//	delta_val	not set for CorrectionStrengths
	correctionstrengths->set_default_properties(correctionstrengths_prop);
	//	Not Polled
	correctionstrengths->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(correctionstrengths);

	//	Attribute : CorrectionStrengthsMSMM
	CorrectionStrengthsMSMMAttrib	*correctionstrengthsmsmm = new CorrectionStrengthsMSMMAttrib();
	Tango::UserDefaultAttrProp	correctionstrengthsmsmm_prop;
	correctionstrengthsmsmm_prop.set_description("CorrectionStrengths spectrum attribute Mean, Std, Min and Max");
	//	label	not set for CorrectionStrengthsMSMM
	//	unit	not set for CorrectionStrengthsMSMM
	//	standard_unit	not set for CorrectionStrengthsMSMM
	//	display_unit	not set for CorrectionStrengthsMSMM
	correctionstrengthsmsmm_prop.set_format("%4.3e");
	//	max_value	not set for CorrectionStrengthsMSMM
	//	min_value	not set for CorrectionStrengthsMSMM
	//	max_alarm	not set for CorrectionStrengthsMSMM
	//	min_alarm	not set for CorrectionStrengthsMSMM
	//	max_warning	not set for CorrectionStrengthsMSMM
	//	min_warning	not set for CorrectionStrengthsMSMM
	//	delta_t	not set for CorrectionStrengthsMSMM
	//	delta_val	not set for CorrectionStrengthsMSMM
	correctionstrengthsmsmm->set_default_properties(correctionstrengthsmsmm_prop);
	//	Not Polled
	correctionstrengthsmsmm->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(correctionstrengthsmsmm);

	//	Attribute : ResonanceStrengths
	ResonanceStrengthsAttrib	*resonancestrengths = new ResonanceStrengthsAttrib();
	Tango::UserDefaultAttrProp	resonancestrengths_prop;
	//	description	not set for ResonanceStrengths
	//	label	not set for ResonanceStrengths
	resonancestrengths_prop.set_unit("m-1");
	//	standard_unit	not set for ResonanceStrengths
	//	display_unit	not set for ResonanceStrengths
	resonancestrengths_prop.set_format("%4.3e");
	//	max_value	not set for ResonanceStrengths
	//	min_value	not set for ResonanceStrengths
	//	max_alarm	not set for ResonanceStrengths
	//	min_alarm	not set for ResonanceStrengths
	//	max_warning	not set for ResonanceStrengths
	//	min_warning	not set for ResonanceStrengths
	//	delta_t	not set for ResonanceStrengths
	//	delta_val	not set for ResonanceStrengths
	resonancestrengths->set_default_properties(resonancestrengths_prop);
	//	Not Polled
	resonancestrengths->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(resonancestrengths);

	//	Attribute : ResonanceStrengthsMSMM
	ResonanceStrengthsMSMMAttrib	*resonancestrengthsmsmm = new ResonanceStrengthsMSMMAttrib();
	Tango::UserDefaultAttrProp	resonancestrengthsmsmm_prop;
	resonancestrengthsmsmm_prop.set_description("ResonanceStrengths spectrum attribute Mean, Std, Min and Max");
	//	label	not set for ResonanceStrengthsMSMM
	//	unit	not set for ResonanceStrengthsMSMM
	//	standard_unit	not set for ResonanceStrengthsMSMM
	//	display_unit	not set for ResonanceStrengthsMSMM
	resonancestrengthsmsmm_prop.set_format("%4.3e");
	//	max_value	not set for ResonanceStrengthsMSMM
	//	min_value	not set for ResonanceStrengthsMSMM
	//	max_alarm	not set for ResonanceStrengthsMSMM
	//	min_alarm	not set for ResonanceStrengthsMSMM
	//	max_warning	not set for ResonanceStrengthsMSMM
	//	min_warning	not set for ResonanceStrengthsMSMM
	//	delta_t	not set for ResonanceStrengthsMSMM
	//	delta_val	not set for ResonanceStrengthsMSMM
	resonancestrengthsmsmm->set_default_properties(resonancestrengthsmsmm_prop);
	//	Not Polled
	resonancestrengthsmsmm->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(resonancestrengthsmsmm);

	//	Attribute : MagnetNames
	MagnetNamesAttrib	*magnetnames = new MagnetNamesAttrib();
	Tango::UserDefaultAttrProp	magnetnames_prop;
	magnetnames_prop.set_description("Level 1 magnet device names");
	//	label	not set for MagnetNames
	//	unit	not set for MagnetNames
	//	standard_unit	not set for MagnetNames
	//	display_unit	not set for MagnetNames
	//	format	not set for MagnetNames
	//	max_value	not set for MagnetNames
	//	min_value	not set for MagnetNames
	//	max_alarm	not set for MagnetNames
	//	min_alarm	not set for MagnetNames
	//	max_warning	not set for MagnetNames
	//	min_warning	not set for MagnetNames
	//	delta_t	not set for MagnetNames
	//	delta_val	not set for MagnetNames
	magnetnames->set_default_properties(magnetnames_prop);
	//	Not Polled
	magnetnames->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(magnetnames);

	//	Attribute : MagnetStates
	MagnetStatesAttrib	*magnetstates = new MagnetStatesAttrib();
	Tango::UserDefaultAttrProp	magnetstates_prop;
	//	description	not set for MagnetStates
	//	label	not set for MagnetStates
	//	unit	not set for MagnetStates
	//	standard_unit	not set for MagnetStates
	//	display_unit	not set for MagnetStates
	//	format	not set for MagnetStates
	//	max_value	not set for MagnetStates
	//	min_value	not set for MagnetStates
	//	max_alarm	not set for MagnetStates
	//	min_alarm	not set for MagnetStates
	//	max_warning	not set for MagnetStates
	//	min_warning	not set for MagnetStates
	//	delta_t	not set for MagnetStates
	//	delta_val	not set for MagnetStates
	magnetstates->set_default_properties(magnetstates_prop);
	//	Not Polled
	magnetstates->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(magnetstates);

	//	Attribute : RequireCycling
	RequireCyclingAttrib	*requirecycling = new RequireCyclingAttrib();
	Tango::UserDefaultAttrProp	requirecycling_prop;
	requirecycling_prop.set_description("Boolean set for each magnets which requires cycling (in beam propagation order)");
	//	label	not set for RequireCycling
	//	unit	not set for RequireCycling
	//	standard_unit	not set for RequireCycling
	//	display_unit	not set for RequireCycling
	//	format	not set for RequireCycling
	//	max_value	not set for RequireCycling
	//	min_value	not set for RequireCycling
	//	max_alarm	not set for RequireCycling
	//	min_alarm	not set for RequireCycling
	//	max_warning	not set for RequireCycling
	//	min_warning	not set for RequireCycling
	//	delta_t	not set for RequireCycling
	//	delta_val	not set for RequireCycling
	requirecycling->set_default_properties(requirecycling_prop);
	//	Not Polled
	requirecycling->set_disp_level(Tango::OPERATOR);
	//	Not Memorized
	att_list.push_back(requirecycling);

	//	Attribute : FrozenMagnets
	FrozenMagnetsAttrib	*frozenmagnets = new FrozenMagnetsAttrib();
	Tango::UserDefaultAttrProp	frozenmagnets_prop;
	frozenmagnets_prop.set_description("List of disabled magnets");
	//	label	not set for FrozenMagnets
	//	unit	not set for FrozenMagnets
	//	standard_unit	not set for FrozenMagnets
	//	display_unit	not set for FrozenMagnets
	//	format	not set for FrozenMagnets
	//	max_value	not set for FrozenMagnets
	//	min_value	not set for FrozenMagnets
	//	max_alarm	not set for FrozenMagnets
	//	min_alarm	not set for FrozenMagnets
	//	max_warning	not set for FrozenMagnets
	//	min_warning	not set for FrozenMagnets
	//	delta_t	not set for FrozenMagnets
	//	delta_val	not set for FrozenMagnets
	frozenmagnets->set_default_properties(frozenmagnets_prop);
	frozenmagnets->set_polling_period(3000);
	frozenmagnets->set_disp_level(Tango::EXPERT);
	//	Not Memorized
	att_list.push_back(frozenmagnets);

	//	Attribute : IsFrozen
	IsFrozenAttrib	*isfrozen = new IsFrozenAttrib();
	Tango::UserDefaultAttrProp	isfrozen_prop;
	isfrozen_prop.set_description("Boolean set for each disabled magnets (in beam propagation order)");
	//	label	not set for IsFrozen
	//	unit	not set for IsFrozen
	//	standard_unit	not set for IsFrozen
	//	display_unit	not set for IsFrozen
	//	format	not set for IsFrozen
	//	max_value	not set for IsFrozen
	//	min_value	not set for IsFrozen
	//	max_alarm	not set for IsFrozen
	//	min_alarm	not set for IsFrozen
	//	max_warning	not set for IsFrozen
	//	min_warning	not set for IsFrozen
	//	delta_t	not set for IsFrozen
	//	delta_val	not set for IsFrozen
	isfrozen->set_default_properties(isfrozen_prop);
	//	Not Polled
	isfrozen->set_disp_level(Tango::EXPERT);
	//	Not Memorized
	att_list.push_back(isfrozen);


	//	Create a list of static attributes
	create_static_attribute_list(get_class_attr()->get_attr_list());
	/*----- PROTECTED REGION ID(MagnetTypeClass::attribute_factory_after) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	MagnetTypeClass::attribute_factory_after
}
//--------------------------------------------------------
/**
 *	Method     : MagnetTypeClass::pipe_factory()
 *	Description: Create the pipe object(s)
 *                and store them in the pipe list
 */
//--------------------------------------------------------
void MagnetTypeClass::pipe_factory()
{
	/*----- PROTECTED REGION ID(MagnetTypeClass::pipe_factory_before) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	MagnetTypeClass::pipe_factory_before
	/*----- PROTECTED REGION ID(MagnetTypeClass::pipe_factory_after) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	MagnetTypeClass::pipe_factory_after
}
//--------------------------------------------------------
/**
 *	Method     : MagnetTypeClass::command_factory()
 *	Description: Create the command object(s)
 *                and store them in the command list
 */
//--------------------------------------------------------
void MagnetTypeClass::command_factory()
{
	/*----- PROTECTED REGION ID(MagnetTypeClass::command_factory_before) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	MagnetTypeClass::command_factory_before


	//	Command Reset
	ResetClass	*pResetCmd =
		new ResetClass("Reset",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pResetCmd);

	//	Command ResetResonanceStrengths
	ResetResonanceStrengthsClass	*pResetResonanceStrengthsCmd =
		new ResetResonanceStrengthsClass("ResetResonanceStrengths",
			Tango::DEV_VOID, Tango::DEV_VOID,
			"",
			"",
			Tango::OPERATOR);
	command_list.push_back(pResetResonanceStrengthsCmd);

	//	Command SetpointCheck
	SetpointCheckClass	*pSetpointCheckCmd =
		new SetpointCheckClass("SetpointCheck",
			Tango::DEVVAR_DOUBLEARRAY, Tango::DEVVAR_STRINGARRAY,
			"List of new strength for each magnet family member",
			"List of magnet device name(s) which refuses their new strength",
			Tango::EXPERT);
	command_list.push_back(pSetpointCheckCmd);

	/*----- PROTECTED REGION ID(MagnetTypeClass::command_factory_after) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	MagnetTypeClass::command_factory_after
}

//===================================================================
//	Dynamic attributes related methods
//===================================================================

//--------------------------------------------------------
/**
 * method : 		MagnetTypeClass::create_static_attribute_list
 * description : 	Create the a list of static attributes
 *
 * @param	att_list	the created attribute list
 */
//--------------------------------------------------------
void MagnetTypeClass::create_static_attribute_list(std::vector<Tango::Attr *> &att_list)
{
	for (unsigned long i=0 ; i<att_list.size() ; i++)
	{
		std::string att_name(att_list[i]->get_name());
		transform(att_name.begin(), att_name.end(), att_name.begin(), ::tolower);
		defaultAttList.push_back(att_name);
	}

	TANGO_LOG_INFO << defaultAttList.size() << " attributes in default list" << std::endl;

	/*----- PROTECTED REGION ID(MagnetTypeClass::create_static_att_list) ENABLED START -----*/
	
	/*----- PROTECTED REGION END -----*/	//	MagnetTypeClass::create_static_att_list
}


//--------------------------------------------------------
/**
 * method : 		MagnetTypeClass::erase_dynamic_attributes
 * description : 	delete the dynamic attributes if any.
 *
 * @param	devlist_ptr	the device list pointer
 * @param	list of all attributes
 */
//--------------------------------------------------------
void MagnetTypeClass::erase_dynamic_attributes(const Tango::DevVarStringArray *devlist_ptr, std::vector<Tango::Attr *> &att_list)
{
	Tango::Util *tg = Tango::Util::instance();

	for (unsigned long i=0 ; i<devlist_ptr->length() ; i++)
	{
		Tango::DeviceImpl *dev_impl = tg->get_device_by_name(((std::string)(*devlist_ptr)[i]).c_str());
		MagnetType *dev = static_cast<MagnetType *> (dev_impl);

		std::vector<Tango::Attribute *> &dev_att_list = dev->get_device_attr()->get_attribute_list();
		std::vector<Tango::Attribute *>::iterator ite_att;
		for (ite_att=dev_att_list.begin() ; ite_att != dev_att_list.end() ; ++ite_att)
		{
			std::string att_name((*ite_att)->get_name_lower());
			if ((att_name == "state") || (att_name == "status"))
				continue;
			std::vector<std::string>::iterator ite_str = find(defaultAttList.begin(), defaultAttList.end(), att_name);
			if (ite_str == defaultAttList.end())
			{
				TANGO_LOG_INFO << att_name << " is a UNWANTED dynamic attribute for device " << (*devlist_ptr)[i] << std::endl;
				Tango::Attribute &att = dev->get_device_attr()->get_attr_by_name(att_name.c_str());
				dev->remove_attribute(att_list[att.get_attr_idx()], true, false);
				--ite_att;
			}
		}
	}
	/*----- PROTECTED REGION ID(MagnetTypeClass::erase_dynamic_attributes) ENABLED START -----*/
	
	/*----- PROTECTED REGION END -----*/	//	MagnetTypeClass::erase_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Method     : MagnetTypeClass::get_attr_object_by_name()
 *	Description: returns Tango::Attr * object found by name
 */
//--------------------------------------------------------
Tango::Attr *MagnetTypeClass::get_attr_object_by_name(std::vector<Tango::Attr *> &att_list, std::string attname)
{
	std::vector<Tango::Attr *>::iterator it;
	for (it=att_list.begin() ; it<att_list.end() ; ++it)
		if ((*it)->get_name()==attname)
			return (*it);
	//	Attr does not exist
	return NULL;
}


/*----- PROTECTED REGION ID(MagnetTypeClass::Additional Methods) ENABLED START -----*/

/*----- PROTECTED REGION END -----*/	//	MagnetTypeClass::Additional Methods
} //	namespace
